import os
import appdirs
import logging
import pandas as pd
from rframe.types import TimeInterval

from .xenon_config import XenonConfig

logger = logging.getLogger(__name__)

from pydantic import BaseSettings

from .clock import SimpleClock

dirs = appdirs.AppDirs("xedocs")

XEDOCS_ENV = os.getenv("XEDOCS_ENV", os.path.join(dirs.user_config_dir, "settings.env"))


class Settings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_"
        env_file = XEDOCS_ENV
        env_file_encoding = "utf-8"

    _DATABASE_INTERFACE_CLASSES = {}
    _RUNDOC_CACHE = {}
    _MONGO_CLIENTS = {}

    CONFIG_DIR = dirs.user_config_dir
    DATA_DIR = dirs.user_data_dir
    DEFAULT_DATABASE = "straxen_db"
    GITHUB_TOKEN: str = None
    
    xenon_config: XenonConfig = XenonConfig()

    clock = SimpleClock()        

    def run_doc(self, run_id, fields=("start", "end")):
        key = (run_id,) + tuple(fields)
        
        if key in self._RUNDOC_CACHE:
            return self._RUNDOC_CACHE[key]
        
        rundb = self.xent_collection()

        if isinstance(run_id, str):
            run_id = int(run_id)

        query = {"number": run_id}

        doc = rundb.find_one(query, projection={f: 1 for f in fields})
        if not doc:
            raise KeyError(f"Run {run_id} not found.")
        
        self._RUNDOC_CACHE[key] = doc

        return doc

    def run_id_to_time(self, run_id):
        doc = self.run_doc(run_id)
        # use center time of run
        time = doc["start"] + (doc["end"] - doc["start"]) / 2
        return self.clock.normalize_tz(time)

    def run_id_to_interval(self, run_id):
        doc = self.run_doc(run_id)
        start = self.clock.normalize_tz(doc["start"]+pd.Timedelta("1s"))
        end = self.clock.normalize_tz(doc["end"]-pd.Timedelta("1s"))
        
        return TimeInterval(left=start, right=end)

    def extract_time(self, kwargs):
        if "time" in kwargs:
            time = kwargs.pop("time")
        elif "run_id" in kwargs:
            time = self.run_id_to_time(kwargs.pop("run_id"))
        else:
            return None
        time = pd.to_datetime(time)
        time = self.clock.normalize_tz(time)
        return time

    def _mongo_client(self, experiment, url=None, user=None, password=None):
        import pymongo

        if experiment not in ["xe1t", "xent"]:
            raise ValueError(
                f"experiment must be 'xe1t' or 'xent'. You passed f{experiment}"
            )

        if not url:
            url = getattr(self.xenon_config.RunDB, f"{experiment}_url")
        if not user:
            user = getattr(self.xenon_config.RunDB, f"{experiment}_user")
        if not password:
            password = getattr(self.xenon_config.RunDB, f"{experiment}_password")

        # build other client kwargs
        max_pool_size = getattr(self.xenon_config.RunDB, "max_pool_size", 100)

        socket_timeout = getattr(self.xenon_config.RunDB, "socket_timeout", 60000)
        connect_timeout = getattr(self.xenon_config.RunDB, "connect_timeout", 60000)

        uri = f"mongodb://{user}:{password}@{url}"
        if uri not in self._MONGO_CLIENTS:
            self._MONGO_CLIENTS[uri] = pymongo.MongoClient(
                uri,
                readPreference="secondaryPreferred",
                maxPoolSize=max_pool_size,
                socketTimeoutMS=socket_timeout,
                connectTimeoutMS=connect_timeout,
            )
        return self._MONGO_CLIENTS[uri]

    def _mongo_database(self, experiment, database=None, **kwargs):
        if not database:
            database = getattr(self.xenon_config.RunDB, f"{experiment}_database")
        client = self._mongo_client(experiment, **kwargs)
        return client[database]

    def _mongo_collection(self, experiment, collection, **kwargs):
        client = self._mongo_database(experiment, **kwargs)
        return client[collection]

    def xent_collection(self, collection="runs", **kwargs):
        return self._mongo_collection("xent", collection, **kwargs)

    def xe1t_collection(self, collection="runs_new", **kwargs):
        return self._mongo_collection("xe1t", collection, **kwargs)

settings = Settings()



from typing import Literal

import rframe

from .base_calibrations import BaseCalibration


class UtubeCalibration(BaseCalibration):
    """Calibrations performed inside the utube"""

    _ALIAS = "utube_calibrations"

    tube: Literal["top", "bottom"] = rframe.Index()
    direction: Literal["cw", "ccw"] = rframe.Index()

    depth_cm: float



"""Main module."""

from pathlib import Path
import pandas as pd

from collections import defaultdict
from typing import Dict, List, Type, Union
from rframe import DataAccessor
from tqdm.auto import tqdm

from ._settings import settings
from .schemas import XeDoc
from .data_locations.mongodb import MongoDB


def find_docs(schema, datasource=None, **labels):
    """find documents by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        List[XeDoc]: a list of documents matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_docs(**labels)


def find_iter(schema, datasource=None, **labels):
    """find documents by labels, return iterator

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        Iterator[XeDoc]: an iterator over documents matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_iter(**labels)


def find_df(schema, datasource=None, **labels):
    """find dataframe by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        DataFrame: a pandas dataframe matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_df(**labels)


def find_one(schema, datasource=None, **labels):
    """find dataframe by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        Optional[XeDoc]: The first document matching selection or None.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_one(**labels)


def insert_docs(schema: str, docs: Union[list, dict, pd.DataFrame], datasource=None, dry=False):
    # Currently stuck on how to deal with instances of schemas
    if datasource == 'straxen_db': # switch to straxen_db
        mongo_username = MongoDB().username
        if mongo_username == 'corrections_expert':
            # If statements only trigger
            target_version = "ONLINE"
            # Note to self: This is done in a very dumb way, fix later
            ONLINE_check = True
            if isinstance(docs, pd.DataFrame):
                ONLINE_check = (docs['version'] == target_version).all()

            elif isinstance(docs, dict):
                ONLINE_check = all(item['version'] == target_version for item in docs)

            else:
                if isinstance(docs, list):
                    if isinstance(docs[0], list) or isinstance(docs[0], dict):
                        # It could be a list of dicts, or list or schemas...
                        # This can be an actual list or a list of docs
                        # these cannot be treated the same
                        ONLINE_check = all(item['version'] == target_version for item in docs)
                    else:
                        # This assumes the last choice is a schema, if not other things will yield errors?
                        # This is kinda sloppy...
                        ONLINE_check = all(item.version == target_version for item in docs)
                elif hasattr(docs, 'version'):
                    if docs.version != target_version: # if version isnt ONLINE
                        ONLINE_check = False
                else:
                    ONLINE_check = all(item.version == target_version for item in docs)

            if not ONLINE_check:
                raise ValueError("You are attempting to modify the a straxen_db correction whose version is not ONLINE")

    if isinstance(docs, pd.DataFrame):
        docs = docs.reset_index().to_dict(orient="records")
    if not isinstance(docs, list):
        docs = [docs]
    accessor = get_accessor(schema, datasource)

    return accessor.insert(docs, dry=dry)


def list_schemas() -> List[str]:
    return list(XeDoc._XEDOCS)


def all_schemas() -> Dict[str, Type[XeDoc]]:
    return dict(XeDoc._XEDOCS)


def schemas_by_category() -> Dict[str, Dict[str, Type[XeDoc]]]:
    d = defaultdict(dict)
    for name, schema in all_schemas().items():
        d[schema._CATEGORY][name] = schema
    return d


def find_schema(name) -> Type[XeDoc]:

    if isinstance(name, type) and issubclass(name, XeDoc):
        return name

    if not isinstance(name, str):
        raise TypeError(
            f"Schema name must be a string or XeDoc class, not {type(name)}"
        )

    schema = XeDoc._XEDOCS.get(name, None)

    if schema is not None:
        return schema

    for schema in XeDoc._XEDOCS.values():
        if schema.__name__ == name:
            return schema

    raise KeyError(f"Correction with name {name} not found.")


def get_accessor(schema, db=None):
    import xedocs

    schema = find_schema(schema)
    if not issubclass(schema, XeDoc):
        raise TypeError(
            "Schema must be a subclass of XeDoc" "or the name of a known schema."
        )

    if db is None:
        db = settings.DEFAULT_DATABASE
    if isinstance(db, str):
        db = getattr(xedocs.databases, db)
    if callable(db):
        db = db()
    return db[schema._ALIAS]


def help(schema):
    if isinstance(schema, str):
        schema = find_schema(schema)
    return schema.help()


def default_datasource_for(schema):
    schema = find_schema(schema)
    accessor = get_accessor(schema)
    return accessor.storage


def get_api_client(schema):
    from .databases import xedocs_api
    schema = find_schema(schema)
    if not issubclass(schema, XeDoc):
        raise TypeError(
            "Schema must be a subclass of XeDoc" "or the name of a known schema."
        )
    db = xedocs_api()
    return db[schema._ALIAS]


def sync_dbs(from_db, to_db, schemas=None, dry=False):
    import xedocs

    if isinstance(from_db, str):
        from_db = getattr(xedocs.databases, from_db)()

    if isinstance(to_db, str):
        to_db = getattr(xedocs.databases, to_db)()

    results = {}
    for name, accessor in tqdm(from_db.items(), desc="Syncing databases"):
        if schemas is None or name in schemas:
            sort=None
            if "time" in accessor.schema.get_index_fields():
                sort = "time"
            docs = accessor.find_docs(sort=sort)
            if name not in to_db:
                continue
            results[name] = to_db[name].insert(docs, raise_on_error=False, dry=dry)

    return results



"""
Correction: Single Electron Gain
Affects: Corrected Areas

Each time the anode is ramped up, there is an exponential relaxation period with a timescale of a day. The electron extraction efficiency and gain are affected, affecting S2 properties that make selection criteria invalid unless corrected. Here is provided the gain to make that correction.
The `ONLINE` version of the correction is used as the average SE gain we want to correct to within straxen. So for any given future run we won't correct for the SE gain in the `ONLINE` processing. Details can be found in the [straxen event processing](https://github.com/XENONnT/straxen/blob/master/straxen/plugins/event_processing.py#L597).
Each local version holds the SE gain at that time, extracted as a per run value.

A useful [note](https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:sr0_ramp_up_kr_se_tudy) by Jianyu.

Wiki:
"""

import rframe

from ..base_corrections import TimeSampledCorrection
from ...constants import PARTITION


class SingleElectronGain(TimeSampledCorrection):

    _ALIAS = "se_gains"
    partition: PARTITION = rframe.Index(default="all_tpc")
    value: float



from .electron_drift import *
from .electron_lifetime import *
from .extraction_efficiency import *
from .field_distortion import *
from .field_simulation import *
from .hit_thresholds import *
from .lone_hits import *
from .pmt_area_to_pe import *
from .position_reconstruction import *
from .s1_aft_xyz import *
from .s1_xyz import *
from .s2_xy import *
from .se_gains import *
from .electron_diffusion_cte import *
from .baseline_samples_nv import *
from .relative_light_yield import *
from .s1_aft_xyz import *
from .avg_se_gain import *
from .bayes_model import *
from .som_network import *
from .hotspot_veto_cut import *
from .photoionization_strength import *
from .s2_pattern_maps import *
from .s1_pattern_maps import *
from .z_bias import *
from .cs2_aft_corrections import *
from .se_gain_partition import *



"""
Correction: Baseline Sample neutron veto
Affects: records_nv

Baseline samples get subtracted from the neutron veto records, similar to the baseline we use for processing recrods. There is a separate array corrections that handles the individual thresholds for each light detector in the Neutron veto

wiki:
"""


from ..base_corrections import TimeSampledCorrection


class BaselineSamplesNV(TimeSampledCorrection):

    _ALIAS = "baseline_samples_nv"
    value: float



#!/usr/bin/env python
"""Tests for `xedocs` package."""
# pylint: disable=redefined-outer-name

import pytest
from click.testing import CliRunner

import xedocs
from xedocs import cli

HAVE_UTILIX = False
try:
    import utilix
    HAVE_UTILIX = True
except ImportError:
    pass


@pytest.mark.skipif(not HAVE_UTILIX, reason='utilix not installed')
def test_straxen_database():
    from xedocs.databases import straxen_db

    db = straxen_db()
    assert db is not None


def test_find_schema():
    from xedocs.schemas import XeDoc

    for name, schema in XeDoc._XEDOCS.items():

        found_schema = xedocs.find_schema(name)
        assert schema is found_schema

        found_schema = xedocs.find_schema(schema.__name__)
        assert schema is found_schema

        found_schema = xedocs.find_schema(schema)
        assert schema is found_schema


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)



import rframe
import requests
from pydantic import BaseSettings
from requests import PreparedRequest
from requests.auth import AuthBase

from ..utils import Database
from .github import GithubCredentials


class ApiAuth(AuthBase):

    def __init__(self, api=None) -> None:
        self.api = api

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        if self.api.token is None:
            self.api.login()
        r.headers["Authorization"] = f"Bearer {self.api.token}"
        return r


class XedocsApi(BaseSettings):
    class Config:
        env_prefix = "xedocs_api_"
    
    base_url: str = "https://api.xedocs.dashboards.xenonnt.org"
    version: str = "v2"
    url_template: str = "{base_url}/{version}/{category}/{name}"
    token: str = None
    
    def fetch_paths(self):
        URL = f"{self.base_url.strip('/')}/{self.version}/.paths"
        headers = {}
        if self.token:
            headers['Authorization'] = f"Bearer {self.token}"
        response = requests.get(URL, headers=headers)
        response.raise_for_status()
        return dict(response.json())
    
    def login(self):
        cred = GithubCredentials.find()
        self.token = cred.token

    def get_datasets(self):
        from xedocs import find_schema

        dsets = {}
        paths = self.fetch_paths()
        for name, path in paths.items():
            schema = find_schema(name)
            url = f"{self.base_url.strip('/')}{path}"
            datasource = rframe.RestClient(
                        url,
                        auth=ApiAuth(self),
                    )
            dsets[name] = rframe.DataAccessor(schema, datasource)
        return Database(dsets)



"""
The 'z_bias_map' calculate the z correction due to non-uniform drift velocity based on
the observed r(r_obs) and z(z_obs).

Reference: xenon:xenonnt:terliuk:drift_field_z_bias_correction
"""

from ..base_references import BaseResourceReference

class ZBias(BaseResourceReference):
    _ALIAS = 'z_bias_maps'
    fmt = 'json.gz'



import datetime
import pandas as pd
import pydantic
import rframe
from rframe.types import TimeInterval
from pydantic import root_validator, validator

from ..base_schemas import XeDoc
from ..._settings import settings


class BaseOperationsReport(XeDoc):
    """Base class for operations report metadata"""

    _ALIAS = ""
    _CATEGORY = "operations"

    class Config:
        allow_population_by_field_name = True

    time: rframe.Interval[datetime.datetime] = rframe.IntervalIndex()

    operator: str = pydantic.Field(min_length=1, max_length=60)
    comments: str

    @validator("time", pre=True)
    def time_string_to_interval(cls, v):
        """Convert str to time interval"""
        if isinstance(v, str):
            try:
                if "," in v:
                    left, right = v.split(",")
                    left, right = pd.to_datetime(left), pd.to_datetime(right)
                    v = TimeInterval(left=left, right=right)
                else:
                    v = pd.to_datetime(v)
            except:
                pass
        return v

    @root_validator(pre=True)
    def run_id_to_time_interval(cls, values):
        if "run_id" in values:
            run_id = values.pop("run_id")
            try:
                values["time"] = settings.run_id_to_interval(run_id)
            except:
                values["time"] = run_id
        return values



"""
# S1 AFT (x,y,z) map
S1 area fraction top data driven map used for event_patternfit within straxen. 
The map is data-driven and depends on the PMT Gains and other observables. 
Should be treated like a normal correction for dependencies and should be updated last for each global version.

"""

from ..base_references import BaseResourceReference


class S1AFTXYZMap(BaseResourceReference):
    _ALIAS = "s1_aft_xyz_maps"
    fmt = "json"

    value: str



import pydantic
import datetime
from typing import Literal, List
from .base_report import BaseOperationsReport


class PmtRate(pydantic.BaseModel):
    pmt: int = pydantic.Field(lt=494, ge=0)
    avg_rate_kbps: float


class HotspotReport(BaseOperationsReport):
    """Hotspot report"""

    _ALIAS = "hotspot_reports"
    severity: Literal["hotspot", "warmspot"] = "hotspot"
    anode_voltage_kv: float
    disppeared_by_itself: bool
    action_taken: str
    plot: str

    pmt_rates: List[PmtRate]



"""
# Correction lone_hits evolution

As studied and discussed in detail in [this note](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:pmt:gains:stability_sr0:lonehits), we could observe a temporal change of the `lone_hits` areas even after correction with the [PMT gains from LED calibrations](https://github.com/XENONnT/corrections/tree/lone_hits_correction/XENONnT/pmt_gains) in XENONnT SR0. As this effect also seems to impact for example the light yield evolution and might be the cause for discrepancies in the Doke-plot, we decided to introduce an emprical correction factor. For simplicity, this correction factor is, for now, the same for all PMTs and based on a piecewise function (23 intervals, constant / linear / exponential), fitted to the median lone_hits evolution for SR0 PMTs and normalized to the stable background data (set to correction factor one). 

The parameters of this correction function are given in the `Lone_hits_correction_SR0.pkl/csv/h5` files (different formats provided in case of incompatibilities). Help and examples on how to apply this correction is given in `Lone_hits_correction_SR0_mockup.ipynb`.


"""



    # represents a correction factor for converting PMT
    # (photomultiplier tube) area to photoelectrons (PE). It is a subclass of
    # `TimeSampledCorrection`, which means that it is a correction that varies with time.
    # The class has fields for `detector`, `pmt`, and `value`. The `detector` field is an
    # index field that specifies the detector (tpc/neutron_veto/muon_veto), while the `pmt` field is an index field
    # that specifies the PMT number. The `value` field is a float that represents the
    # correction factor. (Previously known as gain_models)

import rframe

from ..base_corrections import TimeSampledCorrection
from ...constants import DETECTOR


class PmtAreaToPE(TimeSampledCorrection):
    _ALIAS = "pmt_area_to_pes"

    # Here we use a simple indexer (matches on exact value)
    # to define the pmt field
    # this will add the field to all documents and enable
    # selections on the pmt number. Since this is a index field
    # versioning will be indepentent for each pmt

    detector: DETECTOR = rframe.Index()
    pmt: int = rframe.Index(ge=0)

    value: float



"""
# Position reconstruction with NNs or Likelihoods
Inspired by impressive process in AI industry, we performed deep-learning based approach for position reconstruction. It so far is the fastest reconstruction algorithm, and is less sensitive to the imperfect input pattern. We have a working version of neural network for XENONnT already, but we need someone to optimize it when simulation/data is updated.

In theory, maximum likelihood is the best discriminator for inference. Hence Likelihood-Fitter has always been one of the approach we used to reconstruct position. In XENONnT we will try to improve the likelihood.

See description in the Team C overview page [here](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:reconstruction_team#position_reconstruction_with_neural_network) and [here](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:reconstruction_team#maximum_likelihood_fitter)

Mostly following the "OFF PMTs" list [here](https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:pmt:gains:pmtsoff)


"""

import datetime
from typing import Literal

import rframe

from ..base_references import BaseResourceReference


class PosRecModel(BaseResourceReference):
    _ALIAS = "posrec_models"
    fmt = "binary"

    kind: Literal["cnn", "gcn", "mlp", "s1_cnn"] = rframe.Index()

    value: str
    



import rframe

from .base_pmt_data import BasePmtData


class Pmt(BasePmtData):
    _ALIAS = "pmts"

    serial_number: str = rframe.Index(max_length=60)
    manufacturer: str = rframe.Index(max_length=60)
    location: str
    datasheet: str = ""
    comments: str = ""



"""Top-level package for xedocs."""

__author__ = """Yossi Mosbacher"""
__email__ = "joe.mosbacher@gmail.com"
__version__ = "0.2.27"


import logging

logger = logging.getLogger(__name__)


from ._settings import settings
from . import utils
from . import schemas
from . import data_locations
from . import databases
from .xedocs import *
from .databases import *

try:
    from . import _straxen_plugin
    del _straxen_plugin
except ImportError:
    logger.warning(
        "XEDOCS: Could not register straxen protocol, \
                     most likely straxen not installed."
    )

except ValueError:
    pass

try:
    from . import widgets

    gui = widgets.XedocsEditor()
except ImportError:
    logger.debug("XEDOCS: Could not import editors, GUI not available.")

try:
    from .entrypoints import load_entry_points

    hooks = load_entry_points("xedocs")

    for name, hook in hooks.items():
        if callable(hook):
            hook()
except Exception as e:
    logger.warning(f"XEDOCS: Could not load plugins. Failed with error: {e}")




import os
import configparser

from typing import Any, Dict, Tuple

from pydantic import BaseSettings, BaseModel, SecretStr


def configparser_source(settings: BaseSettings) -> Dict[str, Any]:
    paths = []
    if "HOME" in os.environ:
        paths.append(os.path.join(os.environ["HOME"], ".xenon_config"))

    if "XENON_CONFIG" in os.environ:
        paths.append(os.environ["XENON_CONFIG"])

    paths = [os.path.expanduser(path) for path in paths]

    if not paths:
        return {}
    
    config = configparser.RawConfigParser()
    try:
        config.read(paths)
        return {section: dict(config.items(section)) 
                for section in config.sections() if section in settings.__fields__}
    except:
        return {}


class BasicConfig(BaseModel):
    logging_level = "warning"


class RunDBConfig(BaseModel):
    rundb_api_url: str = ""
    rundb_api_user: str = ""
    rundb_api_password: str = ""
    pymongo_url: str = ""
    pymongo_user: str = ""
    pymongo_password: str = ""
    pymongo_database: str = ""
    xent_url: str = ""
    xent_user: str = ""
    xent_password: str = ""
    xent_database: str = ""
    xe1t_url: str = ""
    xe1t_user: str = ""
    xe1t_password: str = ""
    xe1t_database: str = ""

    max_pool_size: int = 100
    socket_timeout: int = 60000
    connect_timeout: int = 60000
    read_preference: str = "secondaryPreferred"


class StraxenConfig(BaseModel):
    s3_access_key_id: str = ""
    s3_secret_access_key: str = ""


class ScadaConfig(BaseModel):
    sclogin_url: str = ""
    scdata_url: str = ""
    sclastvalue_url: str = ""
    straxen_username: str = ""
    straxen_password: str = ""
    pmt_parameter_names: str = ""


class XenonConfig(BaseSettings):
    basic: BasicConfig = BasicConfig()
    RunDB: RunDBConfig = RunDBConfig()
    straxen: StraxenConfig = StraxenConfig()
    scada: ScadaConfig = ScadaConfig()

    class Config:
        extra = "ignore"
        env_prefix = "XENON_CONFIG_"
        env_nested_delimiter = '__'
        case_sensitive = False

        @classmethod
        def parse_env_var(cls, field_name: str, raw_val: str) -> Any:
            print(f"Parsing {field_name} from {raw_val}")
            return raw_val

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            
            sources = (
                init_settings,
                env_settings,
                configparser_source,
            )
            if os.path.isdir("/run/secrets"):
                sources = sources + (file_secret_settings,)
            return sources
